hazcov  <- function(formula,data,subset,na.action=na.omit,wt,ncg=15,ntg=15,
    failcode=1,minpr=.05,rtime,degree=1,ls=F,eps=.001,max.iter=5,...)
{ 
### Arguments:
###   formula:  a model formula of the form cbind(time,status)~cov1+cov2 (etc)
###     where time is the failure/follow-up time and status is the failure
###     indicator, and cov1 cov2 etc are the covariates to be used in the
###     smoothing.  Although other forms for the covariates can be used,
###     essentially they are always treated as if the full interaction
###     time*cov1*cov2*etc were specified (time should not be specified as a
###     covariate, though, since it is added in as another covariate before
###     the call to loess).  Other functions than cbind() can be used for the
###     response; anything which evaluates to a matrix with time
###     in the first column and status in the second would be ok, as would the
###     name of a matrix with time in the first column and status in the 2nd.
###     In particular, the Surv() function in Terry Therneau's survival 
###     routines could be used.
###   data: optional data frame where formula should be evaluated
###   subset: usual subset argument to modeling functions
###   na.action: usual na.action for modeling functions (see lm), except here
###     the default is set to na.omit.
###   wt: if given, must be ntg by length(time).  (j,i) element gives a prior 
###     relative risk (like an offset) for case i in time interval j.  Default
###     is all 1s.
###   ncg: number of bins to use on the covariate axes.  If one element, the
###     same value is used for all covariates.  If a vector of length
###     ncols(cov), then ntg[i] is the number of bins to use for the ith col of
###     cov. 
###   ntg: number of time intervals
###   failcode: if status==failcode then failtime is a failure time,
###     otherwise failtime is censored
###   minpr: smoothed estimates in the denominator (smoothed follow-up times)
###     are truncated at this value to make sure they dont get too close to 0.
###   rtime: the range on the time axis to be included in the estimation.
###   degree: Degree of the polynomials used in the local regression in loess
###     (1 or 2)
###   ls:  if T, uses iterative local scoring to estimate the log hazard
###     otherwise uses the direct algorithm
###   eps: if ls=T, then the stops when the average relative change in hazard
###     estimates is < eps
###   max.iter:  max number of iterations in the local scoring iteration
###   ...: additional arguments to loess, in particular span, although other
###     arguments (parametric, drop.square, surface, etc) could also be
###     specified. 
###
###  set-up and error checking:
  call <- match.call()
  cov <- match.call(expand = F)
  cov$wt <- cov$ncg <- cov$ntg <- cov$failcode <- cov$mlr <- NULL
  cov$minpr <- cov$rtime <- cov$lrr <- cov$degree <- NULL
  cov$exact.trace <- cov$ls <- cov$eps <- cov$max.iter <- cov$... <- NULL
  cov$na.action <- na.action
  cov[[1]] <- as.name("model.frame.default")
  cov <- eval(cov, sys.parent())
  time <- model.extract(cov,"response")
  status <- time[,2] == failcode
  time <- time[,1]
  if (missing(rtime)) rtime <- c(0,max(time))
###  Not sure if this is foolproof:
  cov <- cov[,as.character(attr(delete.response(terms(cov)),"variables")),
             drop=F]
  ncov <- ncol(cov)
  n <- nrow(cov)
  ncen <- sum(status==0)
  if (missing(wt)) wt <- matrix(1,nrow=ntg,ncol=n)
  else if (nrow(wt) != ntg | ncol(wt) != n) stop("invalid dim(wt)")
  if (length(ncg)==1) ncg <- rep(ncg,ncov)
  else if (length(ncg) != ncov) stop("length(ncg) wrong")
### reduction to grouped data:
### first set up intervals
###    breaks are the interval boundaries, midp the interval midpoints (or the 
###      factor levels), 
###       cov2[i]=k if the jth variable for case i is in
###      interval k.  At end, cell[i] indicates which cell in the grouped
###      covariate space (tensor product of the grid for each covariate) the
###      ith case is in.
  breaks <- NULL
  midp <- NULL
  cell <- rep(0,n)
  for (i in ncov:1) {  
###         catagories for covs which are factors are not altered
###            and these are treated as factor variables by loess
    if (is.factor(cov[,i])) { 
      w <- levels(cov[,i])[table(cov[,i])>0]
      ncg[i] <- length(w)
      cov2 <- as.numeric(factor(cov[,i],levels=w))
      midp <- c(list(factor(w,w)),midp)
    } else {
      w <- sort(unique(cov[,i]))
###  if the number of unique values is <= the intended number of intervals,
###     make each unique value a group.  For truly continuous 
###     covariates the breaks are given by quantiles.
      if (length(w) <= ncg[i]) {
	ncg[i] <- length(w)
	breaks <- c(w[1]-.000001,(w[-ncg[i]]+w[-1])/2,w[ncg[i]],breaks)
	midp <- c(list(w),midp)
	cov2 <- as.numeric(factor(cov[,i],levels=w)) 
      } else {
	w <- sort(unique(c(min(cov[,i])-.000001,quantile(cov[,i],
		       1:(ncg[i]-1)/ncg[i]),max(cov[,i])+.000001)))
	ncg[i] <- length(w)-1
	cov2 <- as.numeric(cut(cov[,i],breaks=w))
	breaks <- c(w,breaks)
	w <- (w[-1]+w[-(ncg[i]+1)])/2
	midp <- c(list(w),midp)
      }
    }
    cell <- ncg[i]*cell+cov2-1
  }
  cell <- cell+1
  celli <- factor(cell)
  cellii <- as.numeric(levels(celli))
  nc <- length(cellii)
  tbrk <- seq(rtime[1],rtime[2],length=ntg+1)
  dt <- diff(tbrk)
  midp <- c(list(tbrk[-1]-dt/2),midp)
  names(midp) <- c("Time",make.names(names(cov),T))
  nl <- nc*ntg   # total number cells
  wt <- ifelse(is.na(wt),0,wt)
  swt <- (wt %*% rep(1,n)) / ((wt>0) %*% rep(1,n))
  o <- order(cell)
  sub <- c(diff(cell[o])>0,T)
  nnn <- rep(0,nc)
  nnn <- table(cell)
  zft <- matrix(0,nrow=ntg,ncol=nc)
  zfs <- zft
  for (i in 1:ntg) {
    ft <- pmin(pmax(time-tbrk[i],0),tbrk[i+1]-tbrk[i])*wt[i,]
    fs <- ifelse(tbrk[i] < time & time <= tbrk[i+1],status,0)
#      zft[i,] <- tapply(ft,cell,sum)
#      zfs[i,] <- tapply(fs,cell,sum)
    ft <- cumsum(ft[o])[sub]
    ft  <- c(ft[1],diff(ft))
    zft[i,] <- ft
    fs <- cumsum(fs[o])[sub]
    fs  <- c(fs[1],diff(fs))
    zfs[i,] <- fs
  }
  zft <- c(zft)
  zfs <- c(zfs)
  dt <- rep(dt,nc)
  nnn <- rep(nnn,rep(ntg,nc))

### set things up and smooth to get fail probs and expected follow-up 
### HD.TMP is the matrix of covariates actually used in the loess smoothing
###   initially it gives all the unique combinations of cov intervals
###   in the same order as components in z (note, expand.grid varies the first
###   argument most rapidly, last argument least rapidly).
  HD.TMP <- expand.grid(midp[-1])[rep(cellii,rep(ntg,nc)),,drop=F]
  HD.TMP <- cbind(Time=rep(midp[[1]],nc),HD.TMP)
  exclude <- zft <= 0
  HD.TMP <- HD.TMP[!exclude,]
### HD.TMP is the data frame for the smoothing, [!exclude] removes cells with
###  no follow-up time, since in the ordinary right censorship case these are
###  beyond the range of the data.
###
  rminpr <- minpr*(rep(swt,nc)[!exclude])
  if (ls) {
    zft <- zft[!exclude]
    zfs <- zfs[!exclude]
    div <- (nnn*dt)[!exclude]
    HD.TMP <- data.frame(mft=zft/div,nf=(zfs-zft)/div,
			 n=nnn[!exclude],HD.TMP,wtt=zft,zfs=zfs)
    form <- paste("loess(mft~",paste(names(midp),collapse="+"),
             ",data=HD.TMP,weights=n,degree=",format(degree),
             ",trace.hat=\"approximate\",...)",sep="")
    form2 <- paste("loess(nf~",paste(names(midp),collapse="+"),
             ",data=HD.TMP,weights=wtt,degree=",format(degree),
             ",...)",sep="")
    a.new <- rep(log(sum(status)/sum(time)),length(zft))
    rss <- 1
    i <- 0
    while(rss > eps & i < max.iter) {
      i <- i+1
      a.old <- a.new
      zftn <- zft*exp(a.old)
      HD.TMP$mft <- zftn/div
      z <- eval(parse(text=form))
      zp1 <- pmax(predict(z),rminpr*zftn/zft)*div
      HD.TMP$wtt <- zp1
      HD.TMP$nf <- a.old+(zfs-zftn)/zp1
      z2 <- eval(parse(text=form2))
      a.new <- predict(z2)
      rss <- sum((a.old-a.new)^2)/sum(a.new^2)
    }
    zp <- exp(a.new)
      aic <- -2*sum(a.new*zfs-zft*zp) +2*z2$inference$trace.hat
      aic <- aic/sum(HD.TMP$n)
  } else {
    HD.TMP <- data.frame(mft=(zft/(nnn*dt))[!exclude],n=nnn[!exclude],
                     nf=(zfs/(nnn*dt))[!exclude],HD.TMP)
### smooth numerator (form2) and denominator (form), and get predicted values
###  for every cell, and estimate hazards by taking the ratio
    form <- paste("loess(mft~",paste(names(midp),collapse="+"),
           ",data=HD.TMP,weights=n,degree=",format(degree),
           ",trace.hat=\"approximate\",...)",sep="")
    form2 <- paste("loess(nf~",paste(names(midp),collapse="+"),
           ",data=HD.TMP,weights=n,degree=",format(degree),
           ",...)",sep="")
    z <- eval(parse(text=form))
    z2 <- eval(parse(text=form2))
    zp1 <- predict(z)
    zp1 <- ifelse(zp1 < rminpr, NA, zp1)
    zp2 <- pmax(predict(z2),0)
    zp <- zp2/zp1
    aic <- NULL
    rss <- NULL
  }
### OUTPUT:
###   call: call to hazcov
###   bin: data frame for smoothing (HD.TMP above) with the estimated hazards 
###      (haz) 
###   fitd: loess output from denominator smooth
###   swt:  n/sum of weights on each time interval
###   fitn: loess output from numerator smooth
###   timeb: boundaries of the time intervals (vector)
###   covb: vector giving boundaries of the covariate intervals 
###   dim: number of intervals for time ([1]) and each cov ([2] to [ncov+1])
###   n: sample size and number censored in the original data (after subset
###      and na.action())
###   midp: list giving the midpoints of the intervals
###   minpr, ls:  input values of these parameters
###   cell: the cell # of the ith case in the binned covariates.  Does not
###      reflect the further splitting of bins for time intervals
###   exclude: bins in the cell x time group tensor product excluded from the
###      rows of bin because of no follow-up time
###   aic: if local scoring used, the value of the aic statistic
###   css: average squared change in estimates at last local scoring iteration
  structure(list(call=call,
      bin=data.frame(haz=zp,HD.TMP),
      fitd=z,swt=swt,fitn=z2,timeb=tbrk,covb=breaks,
      dim=c(ntg,ncg),n=c(sample.size=n,number.censored=ncen),midp=midp,
      minpr=minpr,cell=celli,exclude=exclude,
      ls=ls,aic=aic,css=rss),class="hazcov")
}

print.hazcov <- function(z) {
### z: output from hazcov
  print(z$call)
  print(z$n)
  cat("\nnumber time groups, covariate groups:",format(z$dim))
  cat("\nnumber bins with follow-up time:", format(nrow(z$bin)))
  cat("\nResiduals:\n")
  print(summary(z$bin$nf-z$bin$haz*z$bin$mft))
  cat("\nNumerator Smooth:\n")
  print(z$fitn)
  invisible()
}

predict.hazcov <- function(z,newdata,se.fit=F,lrr=F,baseline,mlr=-log(4),
    		     minpr=z$minpr) { 
###  z: hazcov object
###  newdata: required (z$bin$haz contains hazards at points used in the
###    estimation).  Data frame giving values of time and covariate
###    combinations where estimates should be calculated.  Names in newdata
###    need to correspond to names(z$midp).
###  se.fit: if T, calculate standard errors of the hazard estimates.  If
###    z$bin is large this will often fail due to excessive memory
###    requirements. 
###  lrr: if T, log relative risks are calculated (relative to the covariates
###    specified in baseline)
###  baseline:  list or data.frame containing baseline values for the
###    covariates (attempts to default to a value where estimates are most
###    likely to be defined on all intervals)
###  mlr: if lrr=T, the log relative risks are truncated below at mlr and
###    above by -mlr.  Defaults to value given in z.
###  minpr: estimates only calculated where denominator estimate >= minpr
###    [times a scaling factor if wt used in call to hazcov] (defaults to
###    value given in z). 
###
  assign("HD.TMP",z$bin,1)
#  attr(newdata,"out.attrs")  <- NULL
  zp1 <- predict(z$fitn,newdata=newdata,se.fit=se.fit)
  if (se.fit) {
    zp1s <- c(zp1$se.fit/zp1$residual.scale)
    zp1 <- c(zp1$fit)
  } else {
    zp1 <- c(zp1)
  }
  if (z$ls) {
    zp1 <- exp(zp1) 
  } else {
    zp2 <- c(predict(z$fitd,newdata=newdata))
    zp1 <- pmax(zp1,0)
    ind <- as.numeric(cut(newdata$Time,breaks=z$timeb))
    zp2 <- ifelse(zp2 < minpr*z$swt[ind], NA, zp2)
    zp1 <- zp1/zp2
  }
  zp <- zp1
  if (se.fit) { 
    if (z$ls) { 
      zp1s <- zp1*zp1s # hazard
###      zp1s <- zp1s # log hazard
    } else {
###      zp1s <- ifelse(zp1 > 0, zp1s/sqrt(zp1*zp2), NA) # for log hazard
      zp1s <- sqrt(zp1/zp2)*zp1s   # se of hazard
      ddel <- sqrt(diff(z$timeb)[ind])
      zp1s <- zp1s/ddel
    }
    se <- zp1s
  }     
  if (lrr) {
    if (missing(baseline)) {
      mtx <- max(HD.TMP$Time)
      sub <- HD.TMP$mft == max(HD.TMP$mft[HD.TMP$Time==mtx]) & 
           HD.TMP$Time == mtx
      ind <- match(T,sub)
      baseline <- HD.TMP[ind,names(z$midp)[-1],drop=F]
    } else {
      baseline <- as.data.frame(baseline)
    }
    t1 <- sort(unique(newdata$Time))
    bv <- baseline[rep(1,length(t1)),,drop=F]
    bv <- cbind(Time=t1,bv)
    bv <- bv[,names(z$midp),drop=F]
    zb1 <- c(predict(z$fitn,newdata=bv))
    if (z$ls) {
      zb <- exp(zb1) 
    } else {
      zb2 <- c(predict(z$fitd,newdata=bv))
      zb1 <- pmax(0, zb1)
      ind <- as.numeric(cut(bv$Time,breaks=z$timeb))
      zb2 <- ifelse(zb2 < minpr*z$swt[ind], NA, zb2)
      zb <- zb1/zb2
    }
    index <- match(newdata$Time,t1)
    zb <- zb[index]
    zz <- zp/zb
    zm <- exp(mlr)
    zz <- pmax(zm,zz)
    zz <- log(pmin(1/zm,zz))
    zb <- zz
    if (se.fit) {data.frame(haz=c(zp),logratio=c(zb),se=se)}
    else data.frame(haz=c(zp),logratio=c(zb))
  } else {
    if (se.fit) {data.frame(haz=c(zp),se=se)}
    else c(zp)
  }
}

wcp <- function(z,zb,minpr=zb$minpr,mt=max(z$bin$Time)) {
### z: hazcov output where wcp statistic to be computed
### zb: at risk probabilities estimated from zb for weighted RSS (allows
###   using the same weights with several different hazcov objects)
### minpr: estimates only calculated where at risk probability estimates >
###   minpr (defaults to zb$minpr)
### mt: if z used direct estimation, can restrict the range of times which
###   are included in calculating wcp by giving the maximum as mt
  if (z$ls) {
    z$aic
  } else {
    if (nrow(zb$bin) != nrow(z$bin)) print("data dimension mismatch")
    if (z$fitn$control$trace.hat != "exact") stop(
  "need diagonal of smoother matrix--try refitting with trace.hat=\"exact\"")
    pi <- predict(zb$fitd)
    ind <- as.numeric(cut(zb$bin$Time,breaks=zb$timeb))
    dc <- zb$swt[ind]
    pi <- ifelse(pi < minpr*dc, NA, pi)
    ff <- predict(z$fitn)
    ff <- ifelse(ff>= 0, ff, 0)
    rss <- sum((z$bin$n*((z$bin$nf-ff)/pi)^2)[z$bin$Time <= mt],na.rm=T)
    del <- diff(z$timeb)[ind]
    cc <- sum((z$fitn$inference$diagonal*ff/(pi^2*del))[z$bin$Time <= mt],
            na.rm=T)
    c(rss+2*cc)/sum(z$bin$n[!is.na(pi) & z$bin$Time <= mt])
  }
}

coplot.hazcov <- function(formula,object,given.values,wt,
     cl=F,xlim,ylim,...) {
###  formula must be a valid formula for coplot, eg at most 2 given variables
###  if response starts with an "l" then logratios will be plotted,
###    otherwise hazards will be.
###  three options for given.values:  nothing, include just the given
###    covariates in formula, or include all covariates except the independent
###    variable in formula.  If variables not appearing in formula are not
###    specified in given.values, they are set somewhere near the middle of
###    the corresponding entry in midp.
###  wt (length=ntg) optional vector of prior relative risks.  hazards on time
###    interval i are multiplied by wt[i] 
###  cl:  If T, lines giving +/- 2 standard errors on the log scale added to
###    plots at several points
  if (length(formula)>2) vars <- all.vars(formula)
  else vars <- c("hazard rate",all.vars(formula))
  ncond <- min(length(vars)-2,2)
  if (ncond==1) ngval <- 9 else ngval <- 5
  ind <- match(vars[-1],names(object$midp))
  if (any(is.na(ind))) stop("invalid variable names")
  if (is.factor(object$midp[[ind[1]]])) {
    stop("x must be numeric")
  } else {
    t1 <- list(seq(min(object$midp[[ind[1]]]),max(object$midp[[ind[1]]]),
	           length=40))
  }
  if (missing(given.values)) {
    given.values <- NULL
    for (i in 2:(ncond+1)) { 
      if (is.factor(object$midp[[ind[i]]])) {
	given.values <- c(given.values,object$midp[ind[i]])
      } else {
	given.values <- c(given.values,list(quantile(object$midp[[ind[i]]],
	   probs=(1:ngval)/(ngval+1))))
      }
    }
    names(given.values) <- vars[3:(3+ncond-1)]
  }
  t2 <- NULL
  if (length(given.values) < length(object$midp)-1) {
    t2 <- object$midp[-ind]
    for (i in 1:length(t2)) {
      if (is.factor(t2[[i]])) {
	t2[[i]] <- t2[[i]][1]
      } else {
	t2[[i]] <- median(t2[[i]])
      }
    }
  }
  t1 <- c(t1,given.values,t2)
  names(t1)[1] <- vars[2]
  t2 <- expand.grid(t1)
  if (match(substring(vars[1],1,1),c("l","L"),0)>0) {
    u <- predict(object,t2,lrr=T)[,2]
  } else {
    u <- predict(object,t2,lrr=F)
    if (cl & missing(wt)) {
      t3 <- t1
      t3[[1]] <- t3[[1]][c(4,15,26,37)]
      t3 <- expand.grid(t3)
      uu <- predict(object,t3,lrr=F,se.fit=T)
      ul <- ifelse(uu[,1]>0,exp(log(uu[,1])+2*uu[,2]/uu[,1]),0)
      ll <- ifelse(uu[,1]>0,exp(log(uu[,1])-2*uu[,2]/uu[,1]),0)
      uu <- cbind(c(ul,ll),rbind(t3,t3))
      names(uu)[1] <- vars[1]
    }
    if (!missing(wt)) {
      wtt <- as.numeric(cut(t2$Time,object$timeb))
      u <- u*wt[wtt]
    }
  }
  t2 <- cbind(u,t2[,vars[-1]])
  names(t2)[1] <- vars[1]
  if (missing(xlim)) xlim <- range(t1[[1]],na.rm=T)
  if (!cl | (match(substring(vars[1],1,1),c("l","L"),0)>0) | !missing(wt)){
    if (missing(ylim)) ylim <- range(u,na.rm=T)
    coplot(formula,t2,given.values[vars[-(1:2)]],lines,xlim=xlim,ylim=ylim,...)
  } else {
    if (missing(ylim)) ylim <- range(uu[,1],na.rm=T)
    coplot(formula,t2,given.values[vars[-(1:2)]],lines,xlim=xlim,ylim=ylim,...)
    clf <- function(x,y){
      n2 <- length(y)/2
      segments(x[1:n2],y[1:n2],x[1:n2],y[(n2+1):length(y)])
    }
    coplot(formula,uu,given.values[vars[-(1:2)]],clf,add=T,xlim=xlim,ylim=ylim,...)
  }
  t1[-1]
}

plot.hazcov <- function(object,formula,...) {
### object: output from hazcov
### formula: (optional) formula for call to coplot.hazcov
### ...: other arguments passed to coplot.hazcov
  if (!missing(formula)) {
    out <- coplot.hazcov(formula,object,...)
  } else if (class(object)=="hazcov") {
    lab <- names(object$midp)[-1]
    out <- vector("list",length(lab))
    names(out) <- lab
    for (i in lab) out[[i]] <- coplot.hazcov(as.formula(
		paste("hazard.rate~Time|",i,sep="")),object,...)
  } else {
    stop("first argument must be a hazcov object")
  }
  out
}

persp.hazcov <- function(formula,object,given.values,wt,xlab,ylab,zlab,...) {
###  formula: formula of form resp~x*y.  if resp begins with "l" or "L" then
###    log ratios will be plotted, otherwise hazard rates will be.  x and y
###    are 2 of the variables used in the smoothing (including times)
###  object: a hazcov object
###  given.values: specify values for other covariates (other than x and y in
###    the formula).  If not specified, factors are set to the first level,
###    and continuous variables to somewhere near the middle of the
###    corresponding entries in midp.
###  wt assumed to be hazard ratios (relative to the baseline values)
###    for a given covariate combination at the times in z$midp[[1]]
###    (actually on each of the time intervals)
###  xlab,ylab,zlab: axis labels in call to persp
###  ...: additional arguments passed to persp()
###
  if (length(formula)>2) vars <- all.vars(formula)
  else vars <- c("hazard rate",all.vars(formula))
  ind <- match(vars[-1],names(object$midp))
  if (any(is.na(ind))) stop("invalid variable names")
  if (is.factor(object$midp[[ind[1]]]) | is.factor(object$midp[[ind[2]]]))
    stop ("covariates must be continuous")
  r <- range(object$midp[[ind[1]]],na.rm=T)
  x <- seq(r[1],r[2],length=40)
  r <- range(object$midp[[ind[2]]],na.rm=T)
  y <- seq(r[1],r[2],length=40)
  tmpv <- names(object$midp)[-ind[1:2]]
  if (missing(given.values)) given.values <- NULL
  if (length(tmpv)>0) {
    if (!is.null(given.values)) {
      ind2 <- match(names(given.values),tmpv,0)
      if (any(ind2 == 0)) stop("invalid names in given.values")
      tmpv <- tmpv[-ind2]
    }
    if (length(tmpv)>0) {
      uu <- NULL
      for (i in tmpv) { 
	if (is.factor(object$midp[[i]])) {
	  uu <- c(uu,list(object$midp[[i]][1]))
	} else {
	  uu <- c(uu,list(median(object$midp[[i]])))
	}
      }
      names(uu) <- tmpv
      given.values <- c(given.values,uu)
    }
  }
  uu <- c(list(x=x,y=y),given.values)
  names(uu)[1:2] <- vars[2:3]
  xx <- expand.grid(uu)
  if (match(substring(vars[1],1,1),c("l","L"),0)>0) {
    u <- predict(object,xx,lrr=T)[,2]
    if (missing(zlab)) zlab <- "Log Hazard Ratio"
  } else {
    u <- predict(object,xx,lrr=F)
    if (missing(zlab)) zlab <- "Hazard Rate"
    if (!missing(wt)) {
      wtt <- as.numeric(cut(xx$Time,object$timeb))
      u <- u*wt[wtt]
    }
  }
  if (missing(xlab)) xlab <- names(object$midp)[ind[1]]
  if (missing(ylab)) ylab <- names(object$midp)[ind[2]]
  persp(x,y,matrix(u,nrow=length(x)),xlab=xlab,ylab=ylab,zlab=zlab,...)
  given.values
}

